/** An example Reactor controller for creating thermostat-like schedules
 *  Copyright (c) 2024 Patrick H. Rigney, All Rights Reserved.
 *  WhiteBitController is offered under MIT License - https://mit-license.org/
 *  More info: https://reactor.toggledbits.com/
 */

const version = 24097;

const Controller = require("server/lib/Controller");
const Engine = require("server/lib/Engine").requires( 23339 );

const Logger = require( "server/lib/Logger" );
Logger.getLogger( 'ScheduleController', 'Controller' ).always( "Module ScheduleController v%1", version );

const Configuration = require("server/lib/Configuration");
const logsdir = Configuration.getConfig( "reactor.logsdir" );  /* logs directory path if you need it */

/* Another other modules we need? */
const util = require("server/lib/util");

var impl = false;  /* Implementation data, one copy for all instances, will be loaded by start() later */

var toSSM = function( d ) {
    return d.getHours() * 3600 + d.getMinutes() * 60 + d.getSeconds()
};

module.exports = class ScheduleController extends Controller {
    constructor( struct, id, config ) {
        super( struct, id, config );  /* required */
    }

    /** Start the controller.
     */
    async start() {
        /* Find the sunrise/sunset entity */
        this.suninfo = this.getStructure().findEntity( this.config.sun_entity || "reactor_system>sun" );
        if ( ! this.suninfo ) {
            // Try to locate it by name.
            this.suninfo = this.getStructure().findEntity( "Sun Information" );
        }
        if ( ! this.suninfo ) {
            this.log.warn( `%1 the Reactor System Sun entity could not be located. Please add "sun_entity" to the config for this controller and set it to the canonical ID of the Reactor System Sun Information entity.`,
            this );
        } else if ( ! this.suninfo.hasCapability( "suninfo" ) ) {
            this.log.warn( `%1 the Sun Information (%2) entity is not properly configured; sunrise/sunset edges will not be available.`,
                this, this.suninfo.getCanonicalID() );
            this.suninfo = undefined;
        } else {
            this.log.info( `%1 Sun Information entity is %2`, this, this.suninfo.getCanonicalID() );
        }

        /* Make sure our configuration gives us data to query */
        for ( let eid in ( this.config.entities || {} ) ) {
            let c = this.config.entities[ eid ];
            let e = this.findEntity( eid );
            if ( ! e ) {
                e = this.getEntity( "ValueSensor", eid );
                e.extendCapability( "string_sensor" );
                e.setPrimaryAttribute( "string_sensor.value" );
                e.setName( c.name || eid );
            }
        }

        this.startDelay( 1 );

        this.online();

        /** Must return `this` or a Promise that resolves to this. If this
         *  is declared async (as it should be), the return of a Promise
         *  is implicit (async functions always return Promise).
         */
        return this;
    }

    /** Stop the controller.
     *  You only need to supply this method if your subclass has something
     *  specific it needs to do.
     */
    async stop() {
        this.log.notice( "%1 stopping", this );
        this.stopTimer();
        this.suninfo = null;

        /* Required ending */
        return await super.stop();
    }

    /** run() is called when Controller's single-simple timer expires.
     *  If you don't use Controller's simple timer, you don't need this.
     */
    run() {
        const d = Engine.getInstance().getTime();
        const now = toSSM( d );
        let nextEdge = null;
        for ( let eid in this.config.entities ) {
            this.log.debug( 5, "%1 handling entity %2", this, eid );
            let c = this.config.entities[ eid ];
            let edges = [];
            for ( let edge in ( c.schedule || {} ) ) {
                try {
                    edges.push( { entity: eid, edge: edge, time: this._parse_edge( edge ), value: c.schedule[ edge ] } );
                } catch ( err ) {
                    this.log.err( "%1 entity %2 skipped edge %3: %4", this, eid, edge, err );
                }
            }
            if ( edges.length < 1 ) {
                continue;
            }
            // Sort edges by establishes times for today.
            edges.sort( (a,b) => a.time - b.time );
            // Roll the edge array so that edge[0] is next (>now or wrap around)
            let n = edges.length;  // wrap control
            while ( n-- > 0 && edges[0].time <= now ) {
                edges.push( edges.shift() );
            }
            this.log.debug( 5, "%1 entity %2 now %3 sorted/rolled edges %4", this, eid, now, edges );

            // Active edge is now the last edge in the list.
            const active = edges[edges.length-1];

            // Next edge for this entity is the one in front of us, which is now edges[0] after rolling.
            if ( nextEdge === null || edges[0].time < nextEdge.time ) {
                nextEdge = edges[0];
                this.log.debug( 6, "%1 next edge for scheduler is now %2 %3", this, eid, nextEdge );
            }

            let e = this.getEntity( "ValueSensor", eid );
            if ( e.getAttribute( 'string_sensor.value' ) !== active.value ) {
                this.log.info( "%1 entity %2 edge %3 active, setting primary attribute value to %4", this, e, active.edge, active.value );
                e.deferNotifies( true );
                try {
                    let since = new Date( d );
                    since.setHours( 0, 0, 0 );
                    since.setSeconds( active.time );  // active.time is SSM
                    e.setAttribute( 'string_sensor.value', active.value );
                    e.setAttribute( 'x_schedule_info.edge', active.edge );
                    e.setAttribute( 'x_schedule_info.since', since.getTime() );
                    e.setAttribute( 'x_schedule_info.asof', since.toLocaleTimeString() );
                    e.setAttribute( 'x_schedule_info.edges', JSON.stringify( c.schedule ) );
                } catch ( err ) {
                    this.log.exception( err );
                } finally {
                    e.deferNotifies( false );
                }
            }
        }
        
        // nextEdge, if non-null, is now the earliest next edge for all entities. If null, no more edges today; arm
        // for midnight re-eval.
        this.stopTimer();
        let nd = Engine.getInstance().getTime();
        nd.setHours( Math.floor( nextEdge.time / 3600 ), Math.floor( nextEdge.time / 60 ) % 60, nextEdge.time % 60 );
        if ( nd <= d ) {
            // Next date/time computed before current, go to next day (next edge is at/across midnight).
            nd.setDate( nd.getDate() + 1 );
        }
        this.log.debug( 5, "%1 next event %2 %3 at %4", this, nextEdge.entity, nextEdge.edge, nd.toLocaleString() );

        // Find the delay to the next edge. Rather than just computing seconds straight-away, we DST-proof the
        // calculation by making the Date object compute the difference between the two dates.
        let delay = nd - Engine.getInstance().now();
        this.log.debug( 6, "%1 delay to next event %2 time %3", this, delay, new Date( Engine.getInstance().now() + delay ) );
        this.startDelay( delay );
        this.getSystemEntity().setAttributes({
            next_event:         [ nd.getTime(), nextEdge.entity, nextEdge.edge, nextEdge.value ],
            next_event_date:    nd.toLocaleString()
        }, "x_scheduler");
    }

    _parse_edge( edge ) {
        let m = edge.match( /^\s*(\d+):(\d+)(:(\d+))?\s*$/ );
        if ( m ) {
            let moment = parseInt( m[1] ) * 3600 + parseInt( m[2] ) * 60 + util.coalesce( parseInt( m[4] ), 0 );
            return moment;
        }
        m = edge.toLowerCase().match( /^(sunrise|sunset)\s*/ ); // (\s*(\+|-)\s*(.*))?/ );
        if ( m ) {
            if ( ! this.suninfo ) {
                throw new Error( `The Sun Information entity could not be identified` );
            }
            let ed = new Date( this.suninfo.getAttribute( `suninfo.${m[1]}` ) );
            let moment = toSSM( ed );
            this.log.debug( 6, "%1 edge %2 established %3 at %4 moment %5", this, edge, m[1], ed.toLocaleTimeString(), moment );
            let rest = edge.substr( m[0].length );
            if ( rest.startsWith( '+' ) || rest.startsWith( '-' ) ) {
                let offset = 0;
                let sign = rest.startsWith( '-' ) ? -1 : 1;
                rest = rest.replace( /^(\+|-)\s*/, "" );
                this.log.debug( 6, "%1 edge %2 parsing offset %3", this, edge, rest );
                while ( rest.trim().length ) {
                    let p = rest.match( /^(\d+)\s*(h(ou)?rs?|min(ute)?s?|sec(ond)s?)\s*/ );
                    if ( p ) {
                        this.log.debug( 6, "%1 for %2 handling chunk %3", this, edge, p[0] );
                        let n = parseInt( p[1] );
                        if ( p[2].startsWith( 'h' ) ) {
                            n *= 3600;
                        } else if ( p[2].startsWith( 'm' ) ) {
                            n *= 60;
                        }
                        offset += sign * n;
                        rest = rest.substr( p[0].length );
                    } else {
                        p = rest.match( /\s*(\d+):(\d+)(:(\d+))?/ );
                        if ( p ) {
                            this.log.debug( 6, "%1 for %2 handling chunk %3", this, edge, p[0] );
                            if ( p[3] ) {
                                offset = sign * ( parseInt(p[1]) * 3600 + parseInt(p[2]) * 60 + parseInt(p[4]) );
                            } else {
                                offset = sign * ( parseInt(p[1]) * 3600 + parseInt(p[2]) * 60 );
                            }
                            rest = "";
                        } else {
                            // invalid offset
                            throw new Error( `Offset in "${edge}" unparseable at "${rest}"` );
                            break;
                        }
                    }
                }
                this.log.debug( 6, "%1 final offset for %2 is %3", this, edge, offset );
                moment += offset;
            } else {
                throw new Error( `Offset in "${edge}" expecting +/- at "${rest}"` );
            }
            return moment;
        }
        throw new Error( "Unrecognizable edge format" );
    }

    /** performOnEntity() - we don't have any actions, we're just a data mining
     *  tool (read only), so we don't override the superclass method and let it
     *  handle anything that comes up (which will be nothing).
     */
    /* performOnEntity( entity, actionName, parameters ) { } */

    /** Default action handler for sys_system.restart.
     */
    action_sys_system_restart( entity, params ) {
        this.stopTimer();
        this.startDelay( 1 );
        return Promise.resolve();
    }
};
